package results;

/** Generic constraints for a contest */
public final class Constraints
{
  /** Name of the contest */
  public static final String CONTEST_TITLE = "ContestName";
  /** Maximum points awarded for each contestant */
  public static final int MAX_POINTS = 30;
  /** Minimum points awarded for each contestant */
  public static final int MIN_POINTS = 20;

  /** Compute a histogram of points
   * @param points	points given by a judge
   * @return		a histogram (amounts of points indexed by points)
   */
  public static int[] histogram (int[] points)
  {
    int histogram[] = new int[MAX_POINTS + 1];
    for (int i = points.length; i-- > 0; )
      if (points[i] >= MIN_POINTS && points[i] <= MAX_POINTS)
	histogram[points[i]]++;
    return histogram;
  }

  /** Compute a bitmap of points given too frequently by a judge
   * according to the Standardized Scoring system published in the
   * Organisation Mondiale Coiffure (OMC) Official Jury Briefing
   * Guidelines, September 15, 2003.
   */
  public static boolean[] getFrequent (int[] points)
  {
    final int histogram[] = histogram (points);
    boolean frequent[] = new boolean[MAX_POINTS + 1];
    /* The scores 30, 29 & 28 are never to be duplicated,
       regardless of the number of competitors. */
    frequent[30] = histogram[30] > 1;
    frequent[29] = histogram[29] > 1;
    frequent[28] = histogram[28] > 1;

    if (points.length <= 11) {
      /* Duplication of the score is not allowed.
	 The competitors must be ranked using the scores 20..30. */
      for (int i = MAX_POINTS + 1; i-- > MIN_POINTS; )
	frequent[i] = histogram[i] > 1;
    }
    else if (points.length <= 20) {
      /* The score of 27 can only be given once.
	 The score of 26 can be given a maximum of 2 times.
	 The score of 25 can be given a maximum of 3 times.
	 The scores 24 thru 20 can be duplicated as necessary. */
      frequent[27] = histogram[27] > 1;
      frequent[26] = histogram[26] > 2;
      frequent[25] = histogram[25] > 3;
    }
    else if (points.length <= 30) {
      /* The score of 27 can be giben a maximum of 2 times.
	 The score of 26 can be given a maximum of 3 times.
	 The score of 25 can be given a maximum of 4 times.
	 The scores 24 thru 20 can be duplicated as necessary. */
      frequent[27] = histogram[27] > 2;
      frequent[26] = histogram[26] > 3;
      frequent[25] = histogram[25] > 4;
    }
    else if (points.length <= 50) {
      /* The score of 27 can be giben a maximum of 3 times.
	 The score of 26 can be given a maximum of 4 times.
	 The score of 25 can be given a maximum of 5 times.
	 The scores 24 thru 20 can be duplicated as necessary. */
      frequent[27] = histogram[27] > 2;
      frequent[26] = histogram[26] > 3;
      frequent[25] = histogram[25] > 4;
    }
    else {
      /* The score of 27 can be giben a maximum of 5 times.
	 The score of 26 can be given a maximum of 6 times.
	 The score of 25 can be given a maximum of 7 times.
	 The scores 24 thru 20 can be duplicated as necessary. */
      frequent[27] = histogram[27] > 5;
      frequent[26] = histogram[26] > 6;
      frequent[25] = histogram[25] > 7;
    }

    return frequent;
  }

  /** Compute a bitmap of points given too infrequently by a judge
   * according to the Standardized Scoring system published in the
   * Organisation Mondiale Coiffure (OMC) Official Jury Briefing
   * Guidelines, September 15, 2003.
   */
  public static boolean[] getInfrequent (int[] points)
  {
    final int histogram[] = histogram (points);
    boolean infrequent[] = new boolean[MAX_POINTS + 1];

    /* The juror must use all scoring numbers and must not skip any numbers.
       [Obviously, it is impossible to use all scoring numbers if there are
       less than 11 competitors.  We assume that the requirement
       "must not skip any numbers" must always be followed.] */
    for (int i = MAX_POINTS + 1, n = points.length;
	 i-- > MIN_POINTS && n-- > 0; )
      infrequent[i] = histogram[i] < 1;

    return infrequent;
  }
}
