package results;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Random;
import java.util.TreeSet;
import java.util.Vector;

/** Contest */
public final class Contest
{
  /** Number of privileged jurors in the jury */
  private static final int numPrivilegedJurors = 1;
  /** Series of the contest */
  Series series;
  /** Name of the contest */
  String name;
  /** Jurors for this competition */
  Vector<Judge> judges;
  /** Participants in this competititon */
  Vector<Mirror> mirrors;
  /** Amounts of different mirror numbers (for duplicate detection) */
  private transient Hashtable<Integer,Integer> numbers =
    new Hashtable<Integer,Integer> ();
  /** Picture format (null=no format) */
  private String format;

  /** Flag the data modified */
  void flagModified () {
    series.flagModified ();
    for (Mirror mirror : mirrors)
      mirror.flagModifiedContest ();
  }

  /** Constructor
   * @param name	Name of the contest
   */
  public Contest (String name) {
    this (name, new Vector<Judge> ());
  }

  /** Constructor
   * @param name	Name of the contest
   * @param judges	The jury of this contest
   */
  Contest (String name, Vector<Judge> judges) {
    this.name = name;
    this.judges = judges;
    this.mirrors = new Vector<Mirror> ();
  }

  /** Convert the contest to a printable string */
  public String toString () { return name; }

  /** Get the name of the contest */
  public String getName () { return name; }
  /** Set the name of the contest */
  public void setName (String name) {
    if (!this.name.equals (name)) {
      this.name = name;
      if (series != null) series.flagModified ();
    }
  }
  /** Get the picture format of the contest */
  public String getFormat () { return format; }
  /** Set the picture format of the contest */
  public void setFormat (String format) {
    if (format != null && format.equals (""))
      format = null;
    if (this.format == format)
      return;
    if (this.format != null && format != null &&
	this.format.equals (format))
      return;
    this.format = format;
    if (series != null) series.flagModified ();
  }

  /** Get the number of registered competitors */
  public int getNumMirrors () { return mirrors.size (); }
  /** Get the nth mirror */
  public Mirror getMirror (int i) { return mirrors.elementAt (i); }
  /** Get the mirror of a competitor
   * @param competitor	the competitor
   * @return		the mirror, or null if none registered
   */
  public Mirror getMirror (Competitor competitor) {
    for (Mirror mirror : mirrors)
      if (mirror.competitor == competitor)
	return mirror;
    return null;
  }
  /** Get the number of occurrences of a specific mirror number */
  public int getNumOccurrences (Integer number) {
    if (number == null || number.intValue () == 0) return 0;
    Integer count = numbers.get (number);
    return count == null ? 0 : count.intValue ();
  }
  /** Increment the number of occurrences of a specific mirror number */
  final void incNumbers (int number) {
    if (number == 0) return;
    Integer num = new Integer (number);
    Integer count = numbers.get (num);
    numbers.put (num, new Integer (count == null
				   ? 1
				   : count.intValue () + 1));
  }
  /** Decrement the number of occurrences of a specific mirror number */
  final void decNumbers (int number) {
    if (number == 0) return;
    Integer num = new Integer (number);
    numbers.put (num, new Integer (numbers.get (num).intValue () - 1));
  }

  /** Get the number of registered judges */
  public int getNumJudges () { return judges.size (); }
  /** Get the nth judge */
  public Judge getJudge (int i) { return judges.elementAt (i); }
  /** Swap two judges at specified positions */
  public void swapJudges (int i, int j) {
    Judge judge = judges.elementAt (i);
    judges.setElementAt (judges.elementAt (j), i);
    judges.setElementAt (judge, j);
    flagModified ();
  }
  /** Determine if a judge belongs to the contest */
  public boolean hasJudge (Judge judge) { return judges.contains (judge); }

  /** Register a competitor to the contest
   * @param competitor	Competitor to be registered
   */
  public void add (Competitor competitor) {
    add (new Mirror (competitor, this, 0));
  }
  /** Add a mirror to the contest
   * @param mirror	mirror to be added
   */
  void add (Mirror mirror) {
    if (mirrors.contains (mirror))
      return;
    mirrors.addElement (mirror);
    incNumbers (mirror.number);
    flagModified ();
  }
  /** Register a juror to the contest
   * @param judge	the juror
   */
  public void add (Judge judge) {
    judges.addElement (judge);
    flagModified ();
  }
  /** Remove a judge from the contest
   * @param judge	the judge to be removed
   * @return		true if the judge belonged to the contest
   */
  public boolean remove (Judge judge)
  {
    boolean removed = judges.remove (judge);
    if (removed) flagModified ();
    return removed;
  }
  /** Remove a mirror from the contest
   * @param mirror	the mirror to be removed
   * @return		true if the mirror belonged to the contest
   */
  public boolean remove (Mirror mirror)
  {
    if (!mirrors.remove (mirror))
      return false;
    decNumbers (mirror.number);
    flagModified ();
    return true;
  }
  /** Remove a competitor from the contest
   * @param competitor	the competitor to be removed
   * @return		true if the competitor belonged to the contest
   */
  public boolean remove (Competitor competitor)
  {
    return remove (getMirror (competitor));
  }
  /** Remove all competitors from the contest */
  public void removeCompetitors ()
  {
    if (!mirrors.isEmpty ()) {
      mirrors = new Vector<Mirror> ();
      numbers = new Hashtable<Integer,Integer> ();
      flagModified ();
    }
  }
  /** Add all competitors from the series */
  public void addCompetitors ()
  {
    int i = series.getNumCompetitors ();
    if (i == mirrors.size ())
      return;
    while (i-- > 0)
      add (series.getCompetitor (i));
  }

  /** Shuffle the mirror numbers
   * @param first	lowest mirror number
   */
  public void shuffleMirrors (int first)
  {
    int i;
    Mirror[] nums = new Mirror[mirrors.size ()];
    mirrors.copyInto (nums);
    numbers.clear ();
    for (i = 0; i < nums.length; i++)
      numbers.put (new Integer (first + i), new Integer (1));
    Random random = new Random ();
    while (i > 0) {
      int index = Math.abs (random.nextInt ()) % i;
      Mirror m = nums[--i];
      (nums[i] = nums[index]).number = first + i;
      nums[index] = m;
    }
    for (i = nums.length; i-- > 0; )
      mirrors.setElementAt (nums[i], i);
    flagModified ();
  }

  /** Sort the mirror numbers
   * @param reporter	a reporter object for diagnostics, or null
   */
  public void sortMirrors (ErrorReporter reporter)
  {
    TreeSet<Mirror> sortedMirrors = new TreeSet<Mirror>
      (new MirrorSorter (reporter));
    sortedMirrors.addAll (mirrors);
    mirrors.clear ();
    mirrors.addAll (sortedMirrors);
    flagModified ();
  }

  /** Get an bitmap of points given too frequently by a judge */
  public boolean[] getFrequentPoints (Judge judge)
  {
    int i = mirrors.size ();
    int points[] = new int[i];
    while (i-- > 0) {
      Integer p = mirrors.elementAt (i).getPoints (judge);
      points[i] = p == null ? 0 : p.intValue ();
    }
    return Constraints.getFrequent (points);
  }
  /** Get an bitmap of points given too infrequently by a judge */
  public boolean[] getInfrequentPoints (Judge judge)
  {
    int i = mirrors.size ();
    int points[] = new int[i];
    while (i-- > 0) {
      Integer p = mirrors.elementAt (i).getPoints (judge);
      points[i] = p == null ? 0 : p.intValue ();
    }
    return Constraints.getInfrequent (points);
  }
  /** Determine if a judge has given any points */
  public boolean hasPoints (Judge judge)
  {
    for (int i = mirrors.size (); i-- > 0; )
      if (mirrors.elementAt (i).getPoints (judge) != null)
	return true;
    return false;
  }

  /** Rectify the points given by the jurors */
  public void rectify () {
    for (Mirror mirror : mirrors)
      mirror.rectify ();
  }

  /** Generate a list of the mirror numbers in the competition
   * @param dir		the directory where the list is to be created
   * @param charset	HTML character set
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateMirrors (File dir, String charset,
			       ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "mirrors.html", "../..",
				      series.locale);
    html.header (series.name, name, "mirrors.index", charset);
    boolean printedLine = false;
    TreeSet<Mirror> sortedMirrors = new TreeSet<Mirror>
      (new MirrorSorter (reporter));
    sortedMirrors.addAll (mirrors);
    for (Mirror mirror : sortedMirrors) {
      if (mirror.number == 0) {
	reporter.reportMissingNumber (mirror);
	continue;
      }
      if (!printedLine) {
	printedLine = true;
	html.out.println ("<table class=\"mirror\">\n");
      }
      html.out.println ("<tr>");
      html.out.print ("<td class=\"num\" scope=\"row\">");
      html.out.print (mirror.number);
      html.out.println ("</td>");
      html.out.print ("<td>");
      html.html (mirror.competitor.name);
      html.out.println ("</td>");
      html.out.print ("<td>");
      html.html (mirror.competitor.country.name);
      html.out.println ("</td>");
      html.out.print ("<td>");
      if (mirror.competitor.team != null)
	html.html (mirror.competitor.team.name);
      html.out.println ("</td>");
      html.out.println ("</tr>");
    }
    if (printedLine)
      html.out.println ("</table>\n");
    html.footer ();
  }

  /** Generate individual results for the contest
   * @param dir		the directory where the file is to be created
   * @param charset	HTML character set
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateResults (File dir, String charset,
			       ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "results.html", "../..",
				      series.locale);
    final int jurysize = judges.size ();
    html.header (series.name, name, "results.index", charset);

    html.out.println ("<table class=\"results\" frame=\"hsides\" rules=\"groups\">");
    html.out.println ("<colgroup align=\"right\" span=\"3\">");
    html.out.println ("<colgroup align=\"left\" span=\"2\">");
    html.out.println ("<colgroup align=\"right\" span=\"" + jurysize + "\">");
    html.out.println ("<colgroup align=\"right\" span=\"3\">");
    html.out.println ("<thead><tr>");
    html.tag ("th", "results.rank", "scope=\"col\"");
    html.tag ("th", "results.mirror", "scope=\"col\" colspan=\"2\"");
    html.tag ("th", "results.name", "scope=\"col\"");
    html.tag ("th", "results.team", "scope=\"col\"");
    for (int i = 1; i <= jurysize; i++)
      html.out.println ("<th scope=\"col\"><a href=\"#j" + i + "\">" + i +
			"</a></th>");
    html.tag ("th", "results.total", "scope=\"col\"");
    html.tag ("th", "results.penalty", "scope=\"col\"");
    html.tag ("th", "results.final", "scope=\"col\"");
    html.out.println ("</tr></thead>");

    // flag: has anything been printed?
    boolean output = false;
    // flags: which "cards" have been issued
    boolean yellow = false, rect = false, grey = false;
    // penalty reasons that have occurred
    TreeSet<Reason> penalties = new TreeSet<Reason> (new ReasonSorter ());
    // penalty reason explanations
    Reasons reasons = new Reasons (series.locale);

    // See if a legend needs to be displayed
    for (Mirror mirror : getRanking ()) {
      for (Judge judge : judges) {
	Points points = mirror.getPts (judge);
	if (points == null || points.orig == null)
	  reporter.reportMissingPoints (mirror, judge);
	else {
	  int p = points.orig.intValue ();
	  int average = mirror.rectify ();
	  int rectified = points.rect.intValue ();
	  if (p - average >= 4 || average - p >= 4)
	    yellow = true;
	  else if (!series.national &&
		   judge.country == mirror.competitor.country)
	    grey = true;
	  else if (p != rectified)
	    rect = true;
	}
      }

      if (mirror.penalty != 0) {
	Reason reason = reasons.get
	  ((new Character (mirror.penaltyReason)).toString ());

	if (reason == null)
	  reporter.reportMissingReason (mirror);
	else
	  penalties.add (reason);
      }
    }

    if (yellow || grey || rect || penalties.size () > 0) {
      html.out.println ("<tfoot>");
      if (yellow) {
	html.out.println ("<tr>\n<th scope=\"row\" class=\"points-yellow\"></th>");
	html.tag ("td", "results.yellow.d",
		  "colspan=\"" + (6 + jurysize) + "\"");
	html.out.println ("</tr>");
      }
      if (grey) {
	html.out.println ("<tr>\n<th scope=\"row\" class=\"points-grey\"></th>");
	html.tag ("td", "results.grey.d",
		  "colspan=\"" + (6 + jurysize) + "\"");
	html.out.println ("</tr>");
      }
      if (rect) {
	html.out.println ("<tr>\n<th scope=\"row\" class=\"points-rect\"></th>");
	html.tag ("td", "results.rect.d",
		  "colspan=\"" + (6 + jurysize) + "\"");
	html.out.println ("</tr>");
      }

      for (Reason reason : penalties) {
	html.out.print ("<tr><th scope=\"row\">" + reason.code +
			"</th>\n<td colspan=\"" + (6 + jurysize) + "\">");
	html.html (reason.toString ());
	html.out.println ("</td></tr>");
      }

      html.out.println ("</tfoot>");
    }

    html.out.println ("<tbody>");

    // Display the rankings
    for (Mirror mirror : getRanking ()) {
      output = true;
      html.out.print ("<tr class=\"rect\" id=\"m" + mirror.number +
		      "\">\n <td scope=\"row\" class=\"rank\" rowspan=\"2\">" +
		      mirror.rank +
		      "</td>\n <td class=\"mirrornumber\" rowspan=\"2\">"
		      + mirror.number +
		      "</td>\n <td class=\"mirror\" rowspan=\"2\">");
      if (format != null)
	html.img ("", new Integer (mirror.number).toString (), format);
      html.out.print ("</td>\n <td>");
      html.html (mirror.competitor.name);
      html.out.print ("</td>\n <td>");
      html.html (mirror.competitor.team == null
		 ? mirror.competitor.country.name
		 : mirror.competitor.team.name + ", " +
		 mirror.competitor.country.name);
      html.out.println ("</td>");
      for (Judge judge : judges) {
	Points points = mirror.getPts (judge);
	if (points == null || points.orig == null)
	  html.out.println (" <td class=\"points-none\">-</td>");
	else {
	  int p = points.orig.intValue ();
	  int average = mirror.rectify ();
	  int rectified = points.rect.intValue ();
	  if (p - average >= 4 || average - p >= 4) {
	    html.out.print (" <td class=\"points-yellow\">");
	    yellow = true;
	  }
	  else if (!series.national &&
		   judge.country == mirror.competitor.country) {
	    html.out.print (" <td class=\"points-nat\">");
	    grey = true;
	  }
	  else if (p != rectified) {
	    html.out.print (" <td class=\"points-rect\">");
	    rect = true;
	  }
	  else
	    html.out.print (" <td class=\"points-norm\">");
	  html.out.print (rectified);
	  html.out.println ("</td>");
	}
      }

      html.out.println (" <td class=\"num\">" +
			mirror.getTotalRectifiedNoPenalty () +
			"</td>\n <td class=\"num\">" +
			(mirror.penalty != 0
			 ? (mirror.penalty + " (" + mirror.penaltyReason + ")")
			 : "-") +
			"</td>\n <td class=\"num\">" +
			mirror.getTotalRectified () + "</td>\n</tr>");

      html.out.print ("<tr class=\"orig\">\n <td></td>\n <td></td>\n");
      for (Judge judge : judges) {
	Points points = mirror.getPts (judge);
	if (points == null || points.orig == null)
	  html.out.println (" <td class=\"points-none\"></td>");
	else {
	  int p = points.orig.intValue ();
	  int average = mirror.rectify ();
	  int rectified = points.rect.intValue ();
	  if (p - average >= 4 || average - p >= 4)
	    html.out.print (" <td class=\"points-yellow\">");
	  else if (!series.national &&
		   judge.country == mirror.competitor.country)
	    html.out.print (" <td class=\"points-nat\">");
	  else if (p != rectified)
	    html.out.print (" <td class=\"points-rect\">");
	  else
	    html.out.print (" <td class=\"points-norm\">");
	  if (p != rectified)
	    html.out.print (p);
	  html.out.println ("</td>");
	}
      }

      html.out.println (" <td class=\"num\">" +
			mirror.getTotalNoPenalty () +
			"</td>\n <td class=\"num\"></td>\n" +
			" <td class=\"num\"></td>\n</tr>");
    }

    if (!output)
      html.out.println ("<tr><td></td></tr>"); // thead, tbody mandate tr
    html.out.println ("</tbody>\n</table>\n");

    generateJudges (html);

    html.footer ();
    generateJurors (dir, charset, reporter);
  }

  /** Generate the list of judges for the contest
   * @param html	the HTML output stream
   */
  private void generateJudges (HTMLWriter html)
    throws UnsupportedEncodingException
  {
    html.out.print ("<h3>");
    html.link ("results.judges", "judges.html", null, null);
    html.out.println ("</h3>\n");

    html.out.println ("<table class=\"judges\">\n<tbody>");

    int id = 0;
    for (Judge judge : judges) {
      id++;
      html.out.println ("<tr id=\"j" + id + "\">");
      html.out.print (" <td scope=\"row\">");
      html.link (id, "judges.html", null, "j" + id);
      html.out.print ("</td>\n <td>");
      html.html (judge.name);
      html.out.print ("</td>\n <td>");
      html.html (judge.country.name);
      html.out.println ("</td>\n</tr>");
    }

    if (id == 0)
      html.out.println ("<tr><td></td></tr>"); // thead, tbody mandate tr

    html.out.println ("</tbody>\n</table>\n");
  }

  /** Generate award list for the contest
   * @param html	the HTML output stream
   * @param top		number of top ranks to list
   */
  public void generateAwards (HTMLWriter html, int top)
  {
    // Display the rankings
    for (Mirror mirror : getRanking ()) {
      if (mirror.rank > top)
	break;
      html.out.println ("<p>" + mirror.rank + "." + mirror.competitor.name +
			";" + (mirror.competitor.team == null
			       ? mirror.competitor.country.name
			       : mirror.competitor.team.name + ", " +
			       mirror.competitor.country.name) + "("
			+ mirror.getTotalRectified () + ")</p>");
    }
  }

  /** Generate juror statistics for the contest
   * @param dir		the directory where the file is to be created
   * @param charset	HTML character set
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateJurors (File dir, String charset,
			      ErrorReporter reporter)
    throws IOException, UnsupportedEncodingException
  {
    HTMLWriter html = new HTMLWriter (dir, "judges.html", "../..",
				      series.locale);
    html.header (series.name, name, "judges.index", charset);

    /** Points given indexed by juror */
    Hashtable<Judge,Hashtable<Integer,Vector<Mirror>>> jurors =
      new Hashtable<Judge,Hashtable<Integer,Vector<Mirror>>> ();
    /** Errors indexed by juror */
    Hashtable<Judge,int[]> errors = new Hashtable<Judge,int[]> ();
    /** All different points given by different judges */
    Vector<Integer> givenPoints = new Vector<Integer> ();

    // Gather the statistics
    for (Judge judge : judges) {
      /** Mirrors indexed by points given */
      Hashtable<Integer,Vector<Mirror>> points =
	new Hashtable<Integer,Vector<Mirror>> ();
      /** Number of yellow cards (rectifications), not counting "grey cards" */
      int numYellow = 0;
      /** Total effect of all rectifications */
      int totalRectified = 0;
      for (Mirror mirror : mirrors) {
	Points p = mirror.getPts (judge);
	if (p != null && p.orig != null) {
	  /* Gather rectification statistics */
	  int average = mirror.computeRectified ();
	  int diff = Math.abs (p.orig.intValue () - average);
	  if (diff >= 4)
	    numYellow++;
	  totalRectified += diff;
	  if (!givenPoints.contains (p.orig)) {
	    int i;
	    for (i = 0; i < givenPoints.size (); i++) {
	      if (givenPoints.elementAt (i).intValue () < p.orig.intValue ()) {
		givenPoints.insertElementAt (p.orig, i);
		break;
	      }
	    }
	    if (i == givenPoints.size ())
	      givenPoints.addElement (p.orig);
	  }
	  Vector<Mirror> pointmirrors = points.get (p.orig);
	  if (pointmirrors == null)
	    points.put (p.orig, pointmirrors = new Vector<Mirror> ());
	  pointmirrors.addElement (mirror);
	}
      }
      jurors.put (judge, points);
      errors.put (judge, new int[] {numYellow, totalRectified});
    }

    // Display the statistics
    html.out.println ("<table frame=\"hsides\" rules=\"groups\">");
    html.out.println ("<colgroup align=\"left\" span=\"3\">");
    html.out.println ("<colgroup align=\"left\" span=\"" +
		      givenPoints.size() + "\">");
    html.out.println ("<thead>\n<tr>\n <th></th>\n <th></th>\n <th></th>");
    for (Integer p : givenPoints)
      html.out.println (" <th scope=\"col\">" + p + "</th>");
    html.out.println ("</tr>\n<tr>");
    html.tag ("th", "judges.name", "scope=\"col\" rowspan=\"2\"");
    html.tag ("th", "judges.yellow1", "scope=\"col\"");
    html.tag ("th", "judges.total1", "scope=\"col\"");
    for (Integer p : givenPoints)
      html.tag ("th", "judges.mirror1", "scope=\"col\"");
    html.out.println ("</tr>\n<tr>");
    html.tag ("th", "judges.yellow2", "scope=\"col\"");
    html.tag ("th", "judges.total2", "scope=\"col\"");
    for (Integer p : givenPoints)
      html.tag ("th", "judges.mirror2", "scope=\"col\"");
    html.out.println ("</tr>\n</thead>\n<tbody>");

    final String span = givenPoints.size () > 0 ? " rowspan=\"2\"" : "";
    /** Index number of each judge */
    int id = 0;

    for (Judge judge : judges) {
      Hashtable<Integer,Vector<Mirror>> points = jurors.get (judge);
      int err[] = errors.get (judge);

      html.out.println ("<tr id=\"j" + (++id) +
			"\">\n <td scope=\"row\"" + span + ">" + judge.name +
			"</td>\n <td class=\"num\"" + span + ">" + err[0] +
			"</td>\n <td class=\"num\"" + span +">" + err[1] +
			"</td>");

      for (Integer p : givenPoints) {
	Vector<Mirror> pmirrors = points.get (p);
	if (pmirrors != null) {
	  html.out.print (" <td>");
	  for (Enumeration<Mirror> m = pmirrors.elements ();
	       m.hasMoreElements (); ) {
	    Mirror mirror = m.nextElement ();
	    html.out.print (mirror.rank);
	    if (m.hasMoreElements ())
	      html.out.print (",");
	  }
	  html.out.print ("</td>");
	}
	else
	  html.out.println (" <td></td>");
      }

      html.out.println ("</tr>");

      if (givenPoints.size () > 0) {
	html.out.println ("<tr>");

	for (Integer p : givenPoints) {
	  Vector<Mirror> pmirrors = points.get (p);
	  if (pmirrors != null) {
	    html.out.print (" <td>");
	    for (Enumeration<Mirror> m = pmirrors.elements ();
		 m.hasMoreElements (); ) {
	      Mirror mirror = m.nextElement ();
	      html.link (mirror.number, "results.html",
			 null, "m" + mirror.number);
	      if (m.hasMoreElements ())
		html.out.print (",");
	    }
	    html.out.println ("</td>");
	  }
	  else
	    html.out.println (" <td></td>");
	}

	html.out.println ("</tr>");
      }

      // Complain about frequent and infrequent points
      boolean[] freq = getFrequentPoints (judge);
      for (int i = 0; i < freq.length; i++)
	if (freq[i])
	  reporter.reportFrequent (i, judge, this);
      freq = getInfrequentPoints (judge);
      for (int i = 0; i < freq.length; i++)
	if (freq[i])
	  reporter.reportInfrequent (i, judge, this);
    }
    if (id == 0)
      html.out.println ("<tr><td></td></tr>"); // thead, tbody mandate tr
    html.out.println ("</tbody>\n</table>");
    html.footer ();
  }

  /** Determine the rankings
   * @return	Mirrors in the ranking order
   */
  List<Mirror> getRanking ()
  {
    List<Mirror> ranking = new ArrayList<Mirror> (mirrors);
    final Collator collator = Collator.getInstance (series.locale);
    Collections.sort (ranking, new Comparator<Mirror> () {
	public int compare (Mirror a, Mirror b) {
	  int aTotal = a.getTotalRectified ();
	  int bTotal = b.getTotalRectified ();
	  if (aTotal != bTotal)
	    return bTotal - aTotal;
	  if (judges.size () > 0) {
	    Integer aPoints[] = a.getPoints ();
	    Integer bPoints[] = b.getPoints ();
	    for (int j = 0; j < numPrivilegedJurors; j++) {
	      if (aPoints[j] == null || bPoints[j] == null)
		continue;
	      aTotal = aPoints[j].intValue () * aPoints.length - a.penalty;
	      bTotal = bPoints[j].intValue () * bPoints.length - b.penalty;
	      if (aTotal != bTotal)
		return bTotal - aTotal;
	    }
	  }

	  // Still even game: compare number of highest points
	  PointJudges apj[] = a.getHistogram ();
	  PointJudges bpj[] = b.getHistogram ();

	  for (int j = 0; j < apj.length && j < bpj.length; j++) {
	    aTotal = apj[j].points;
	    bTotal = bpj[j].points;
	    if (aTotal != bTotal)
	      return bTotal - aTotal;
	    aTotal = apj[j].getNumJudges ();
	    bTotal = bpj[j].getNumJudges ();
	    if (aTotal != bTotal)
	      return bTotal - aTotal;
	  }

	  // It is an even game: collate the competitors in alphabetical order
	  int cmp = collator.compare (a.competitor.name, b.competitor.name);
	  if (cmp < 0)
	    b.evengame = true;
	  else if (cmp > 0)
	    a.evengame = true;
	  return cmp;
	}
      });

    int place = 0, rank = 0;
    for (Mirror mirror : ranking) {
      if (!mirror.evengame)
	rank = place + 1;
      mirror.rank = rank;
      place++;
    }

    return ranking;
  }
}
