package results;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.Locale;
import java.util.ResourceBundle;
import java.text.MessageFormat;

/** HTML utility class */
public final class HTMLWriter
{
  /** The resource bundle for localized strings */
  private ResourceBundle bundle;
  /** The message format for formatted output */
  private MessageFormat mf;
  /** Path name of the top-level directory, e.g., ".." or "../..", or null */
  private final String top;
  /** The character encoding of the HTML document */
  private String encoding;
  /** The output stream */
  PrintWriter out;

  /** Constructor
   * @param dir		the output directory
   * @param name	the file name label
   * @param top		top level directory, e.g., ".." or "../..", or null
   * @param locale	the locale
   * @exception		IOException if the output file cannot be created
   */
  public HTMLWriter (File dir, String name, String top, Locale locale)
    throws IOException
  {
    bundle = ResourceBundle.getBundle ("html", locale);
    (mf = new MessageFormat ("")).setLocale (locale);
    this.top = top;
    out = new PrintWriter (new FileWriter
      (new File (dir, bundle.getString (name))));
  }

  /** Called by the garbage collector before the object is discarded */
  protected void finalize ()
  {
    if (out != null)
      footer ();
  }

  /** Generate a translation of a label
   * @param label	the label
   */
  public void translate (String label)
  {
    out.print (bundle.getString (label));
  }

  /** Format a translation of a label
   * @param label	the label
   * @param params	the parameters to format
   */
  public void format (String label, Object[] params)
  {
    mf.applyPattern (bundle.getString (label));
    out.print (mf.format (params));
  }

  /** Look up a translation of a label
   * @param label	the label
   * @return		the translation
   */
  public String lookup (String label)
  {
    return bundle.getString (label);
  }

  /** Generate an HTML header
   * @param title	Main title
   * @param title2	Secondary title (displayed after "&mdash;" if non-null)
   * @param subtitle	Subtitle (translated if non-null)
   * @param charset	HTML character set (or null)
   */
  public void header (String title, String title2,
		      String subtitle, String charset)
  {
    out.print ("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n\n<html><head>\n");
    encoding = charset;
    if (encoding == null)
      encoding = "utf-8";

    out.print ("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=");
    out.print (encoding);
    out.print ("\">\n");

    if (top != null) {
      out.print ("<link rel=\"stylesheet\" media=\"screen\" href=\"");
      out.print (top);
      out.println ("/screen.css\">");
      out.print ("<link rel=\"stylesheet\" media=\"print\" href=\"");
      out.print (top);
      out.println ("/print.css\">");
    }
    out.print ("<title>");
    html (title);
    if (title2 != null) {
      out.print ("&mdash;");
      html (title2);
    }
    out.print ("</title>\n</head>\n<body>\n\n<h1>");
    html (title);
    if (title2 != null) {
      out.print ("&mdash;");
      html (title2);
    }
    out.println ("</h1>");

    if (subtitle != null) {
      out.print ("<h2>");
      translate (subtitle);
      out.println ("</h2>");
    }
  }

  /** Generate an HTML footer */
  public void footer ()
  {
    out.println ("</body></html>");
    out.close ();
    out = null;
  }

  /** Generate a HTML tag with translated content
   * @param tag		the tag name
   * @param label	the label of the content
   * @param attr	the attributes of the element
   */
  public void tag (String tag,
		   String label,
		   String attr)
  {
    out.print ("<");
    out.print (tag);
    if (attr != null) {
      out.print (" ");
      out.print (attr);
    }
    out.print (">");
    translate (label);
    out.print ("</");
    out.print (tag);
    out.println (">");
  }

  /** Generate markup for a link
   * @param name	label of the link name (null=use directory name)
   * @param target	label of the link target
   * @param directory	link target directory name (untranslated, null=none)
   * @param subtarget	link subtarget name (untranslated, null=none)
   */
  public void link (String name,
		    String target,
		    String directory,
		    String subtarget)
    throws UnsupportedEncodingException
  {
    out.print ("<a href=\"");
    if (directory != null) {
      url (directory);
      out.print ('/');
    }
    url (bundle.getString (target));
    if (subtarget != null) {
      out.print ('#');
      url (subtarget);
    }
    out.print ("\">");
    if (name != null)
      translate (name);
    else
      html (directory);
    out.print ("</a>");
  }

  /** Generate markup for a link
   * @param name	the link name
   * @param target	label of the link target
   * @param directory	link target directory name (untranslated, null=none)
   * @param subtarget	link subtarget name (untranslated, null=none)
   */
  public void link (int name,
		    String target,
		    String directory,
		    String subtarget)
    throws UnsupportedEncodingException
  {
    out.print ("<a href=\"");
    if (directory != null) {
      url (directory);
      out.print ('/');
    }
    url (bundle.getString (target));
    if (subtarget != null) {
      out.print ('#');
      url (subtarget);
    }
    out.print ("\">");
    out.print (name);
    out.print ("</a>");
  }

  /** Generate markup for an inlined image
   * @param directory	the directory name
   * @param filename	name of the image file
   */
  public void img (String directory, String filename)
    throws UnsupportedEncodingException
  {
    out.print ("<a href=\"");
    if (!directory.equals ("")) {
      url (directory);
      out.print ('/');
    }
    url (filename);
    out.print ("\"><img alt=\"");
    translate ("picture");
    out.print ("\" src=\"");
    if (!directory.equals ("")) {
      url (directory);
      out.print ('/');
    }
    url ("t/");
    url (filename);
    out.print ("\"></a>");
  }

  /** Generate markup for inlined images
   * @param directory	the directory name
   * @param basename	base name of the image file
   * @param suffixes	suffixes of the image files, separated by spaces
   */
  public void img (String directory, String basename, String suffixes)
    throws UnsupportedEncodingException
  {
    int i = 0, j = suffixes.indexOf (' ');
    while (j >= 0) {
      img (directory, basename + suffixes.substring (i, j));
      i = j + 1;
      j = suffixes.indexOf (' ', i);
    }
    if (suffixes.length () > i)
      img (directory, basename + suffixes.substring (i));
  }

  /** Output a URL string
   * @param name	URL to be output
   */
  public void url (String name) throws UnsupportedEncodingException
  {
    out.print (urlencode (name));
  }

  /** Output an HTML string
   * @param name	text to be output
   */
  public void html (String text)
  {
    out.print (htmlencode (text));
  }

  /** Encode a name as a URI
   * @param name	name to be encoded
   * @return		the name converted to a URI
   */
  public String urlencode (String name) throws UnsupportedEncodingException
  {
    return URLEncoder.encode (name, encoding).
      replace ("+", "%20").replace ("%2F", "/");
  }

  /** Encode a name as a HTML string
   * @param name	name to be encoded
   * @return		the name converted to HTML text
   */
  public static String htmlencode (String name)
  {
    StringBuffer encoded = new StringBuffer (name.length ());
    for (int i = 0; i < name.length (); i++) {
      switch (name.charAt (i)) {
      case '&':
	encoded.append ("&amp;");
	break;
      case '<':
	encoded.append ("&lt;");
	break;
      case '>':
	encoded.append ("&gt;");
	break;
      default:
	encoded.append (name.charAt (i));
      }
    }
    return encoded.toString ();
  }
}
