package results;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.TreeMap;

/** Mirror in a contest */
public final class Mirror
{
  /** The competitor */
  Competitor competitor;
  /** The contest */
  Contest contest;
  /** The points, indexed by the judges */
  private Hashtable<Judge,Points> pt = new Hashtable<Judge,Points> ();
  /** The mirror number (0=undefined) */
  int number;
  /** Penalty points */
  int penalty = 0;
  /** Primary reason for giving penalty points */
  char penaltyReason = 0;

  /** Average points (cached) */
  private transient int points = 0;
  /** Total points (cached) */
  private transient int total = 0;
  /** Average points, ignoring some judges (cached) */
  transient int ipoints = 0;
  /** Total rectified points (cached) */
  private transient int itotal = 0;
  /** Total rectified points (cached) */
  private transient int iitotal = 0;
  /** Rank (cached) */
  transient int rank = 0;
  /** Flag: is this an even game with the previous mirror? */
  transient boolean evengame = false;

  /** Constructor
   * @param competitor	The competitor
   * @param contest	The contest
   * @param number	The mirror number
   */
  public Mirror (Competitor competitor, Contest contest, int number) {
    this.competitor = competitor;
    this.contest = contest;
    this.number = number;
  }

  /** Get the contest */
  public Contest getContest () { return contest; }
  /** Get the competitor */
  public Competitor getCompetitor () { return competitor; }
  /** Set the competitor */
  public void setCompetitor (Competitor competitor) {
    this.competitor = competitor;
    contest.series.flagModified ();
  }
  /** Get the mirror number as an integer */
  public int getNum () { return number; }
  /** Get the mirror number */
  public Integer getNumber () {
    return number != 0 ? new Integer (number) : null;
  }
  /** Set the mirror number */
  public void setNumber (Integer number) {
    contest.decNumbers (this.number);
    this.number = number == null ? 0 : number.intValue ();
    contest.incNumbers (this.number);
    contest.series.flagModified ();
  }
  /** Get the penalty points */
  public Integer getPenalty () {
    return penalty != 0 ? new Integer (penalty) : null;
  }
  /** Set the penalty points */
  public void setPenalty (Integer penalty) {
    this.penalty = penalty == null ? 0 : penalty.intValue ();
    contest.series.flagModified ();
  }
  /** Get the penalty reason */
  public char getPenaltyReason () {
    return penaltyReason;
  }
  /** Set the penalty reason */
  public void setPenaltyReason (Reason reason) {
    penaltyReason = reason == null ? 0 : reason.code;
    contest.series.flagModified ();
  }

  /** Relaxed equality comparison
   * @param object	object to be compared to this
   * @return		true if the competitors match
   */
  public boolean equals (Object object) {
    return this == object ||
      (object instanceof Competitor && competitor == object) ||
      (object instanceof Mirror && competitor == ((Mirror) object).competitor);
  }

  /** Change the points given to this mirror
   * @param judge	The judge
   * @param points	The points
   * @exception		IllegalArgumentException if the judge is invalid
   */
  public void setPoints (Judge judge, Integer points) {
    if (!contest.judges.contains (judge))
      throw new IllegalArgumentException ();
    pt.put (judge, new Points (judge, points));
    flagModified ();
  }

  /** Get the points given to this mirror
   * @param points	a table of points in the jury order
   * @exception		IllegalArgumentException on table length mismatch
   */
  void setPoints (Integer[] points) {
    int i = contest.judges.size ();
    if (i != points.length)
      throw new IllegalArgumentException ();
    while (i-- > 0) {
      Judge judge = contest.judges.elementAt (i);
      pt.put (judge, new Points (judge, points[i]));
    }
    flagModified ();
  }

  /** Get the total original points for this mirror, minus penalty */
  public int getTotal () {
    computePoints ();
    return total - penalty;
  }
  /** Get the total original points for this mirror */
  public int getTotalNoPenalty () {
    computePoints ();
    return total;
  }
  /** Get the total rectified points for this mirror, minus penalty */
  public int getTotalRectified () {
    if (iitotal == 0) rectify ();
    return iitotal - penalty;
  }
  /** Get the total rectified points for this mirror */
  public int getTotalRectifiedNoPenalty () {
    if (iitotal == 0) rectify ();
    return iitotal;
  }

  /** Flag the mirror data modified */
  private void flagModified () {
    contest.flagModified ();
    points = ipoints = 0;
    total = itotal = iitotal = 0;
  }

  /** Flag the contest data modified */
  void flagModifiedContest () {
    rank = 0;
    evengame = false;
    points = ipoints = 0;
    total = itotal = iitotal = 0;
    for (Points p : pt.values ())
      p.rect = null;
  }

  /** Determine if any points have been assigned to the mirror */
  public boolean hasPoints () {
    for (int i = contest.judges.size (); i-- > 0; )
      if (pt.get (contest.judges.elementAt (i)) != null)
	return true;
    return false;
  }
  /** Get the points given to this mirror
   * @return		a table of points in the jury order
   */
  public Integer[] getPoints () {
    int i = contest.judges.size ();
    Integer result[] = new Integer[i];
    while (i-- > 0) {
      Points pts = pt.get (contest.judges.elementAt (i));
      result[i] = pts != null ? pts.orig : null;
    }
    return result;
  }

  /** Get the points given to this mirror by a judge
   * @param judge	The judge
   * @return		The points
   * @exception		IllegalArgumentException if the judge is invalid
   */
  public Integer getPoints (Judge judge) {
    final Points pts = getPts (judge);
    return pts == null ? null : pts.orig;
  }

  /** Get the points given to this mirror by a judge
   * @param judge	The judge
   * @return		The points
   * @exception		IllegalArgumentException if the judge is invalid
   */
  Points getPts (Judge judge) {
    if (!contest.judges.contains (judge))
      throw new IllegalArgumentException ();
    return pt.get (judge);
  }

  /** Get the rectified points given to this mirror
   * @return		a table of points in the jury order
   */
  public Integer[] getRectified () {
    int i = contest.judges.size ();
    Integer result[] = new Integer[i];
    while (i-- > 0) {
      Points pts = pt.get (contest.judges.elementAt (i));
      result[i] = pts != null ? pts.rect : null;
    }
    return result;
  }

  /** Get the point distribution given to the mirror
   * @return	an array of points and a list of judges who give those points
   */
  public PointJudges[] getHistogram () {
    TreeMap<Integer,PointJudges> histogram =
      new TreeMap<Integer,PointJudges> ();
    int i = contest.judges.size ();
    while (i-- > 0) {
      Judge judge = contest.judges.elementAt (i);
      Points pts = pt.get (judge);
      if (pts != null && pts.orig != null) {
	Integer p = pts.orig;
	PointJudges pj = histogram.get (p);
	if (pj == null) {
	  pj = new PointJudges (p.intValue ());
	  histogram.put (p, pj);
	}
	pj.add (judge);
      }
    }
    PointJudges pj[] = new PointJudges[i = histogram.size ()];
    for (PointJudges pji : histogram.values ())
      pj[--i] = pji;
    return pj;
  }

  /** Compute the average points given for this mirror */
  private void computePoints () {
    if (points != 0)
      return;
    int count = 0, icount = 0;
    total = itotal = iitotal = 0;
    for (Judge judge : contest.judges) {
      Points pts = pt.get (judge);
      if (pts != null && pts.orig != null) {
	int i = pts.orig.intValue ();
	total += i;
	count++;
	if (!contest.series.national &&
	    judge.country == competitor.country)
	  continue; // rule out the judge
	itotal += i;
	icount++;
      }
    }
    points = count > 0 ? (total + count / 2) / count : 0;
    ipoints = icount > 0 ? (itotal + icount / 2) / icount : 0;
  }

  /** Compute the average rectified points given for this mirror
   * @return	the average rectified points
   */
  int computeRectified ()
  {
    int count = 0;
    itotal = iitotal = 0;
    for (Judge judge : contest.judges) {
      Points pts = pt.get (judge);
      if (pts != null) {
	if (pts.rect == null) {
	  pts.rect = pts.orig;
	  if (pts.rect == null)
	    continue;
	}
	iitotal += pts.rect.intValue ();
	if (contest.series.national ||
	    judge.country != competitor.country) {
	  itotal += pts.rect.intValue ();
	  count++;
	}
      }
    }
    return count > 0 ? (itotal + count / 2) / count : 0;
  }

  /** Rectify the points for this mirror
   * @return	the average points
   */
  public int rectify () {
    computePoints ();
    int rectified = ipoints;
    // exclude judges from the same country
    if (!contest.series.national) {
      for (Judge judge : contest.judges)
	if (judge.country == competitor.country)
	  pt.get (judge).rectify0 (rectified);
      // finally, update "itotal"
      rectified = computeRectified ();
    }
    // Replace points deviating more than 3 points from the average
    // with the average.
    for (Judge judge : contest.judges) {
      Points pts = pt.get (judge);
      if (pts == null)
	pt.put (judge, new Points (judge, null));
      else
	pts.rectify3 (rectified);
    }
    rectified = computeRectified ();

    // Replace points deviating more than 2 points from the average
    // with average-2 or average+2.
    for (Judge judge : contest.judges) {
      Points pts = pt.get (judge);
      if (pts == null)
	pt.put (judge, new Points (judge, null));
      else
	pts.rectify2 (rectified);
    }

    rectified = computeRectified ();
    return rectified;
  }
}
