package results;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.Writer;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;
import java.util.Vector;

/** Series of contests */
public final class Series
{
  /** The file name extension for series files */
  public final static String FILE_EXTENSION = ".pts";
  /** The file the series was last synchronized with */
  public File file;
  /** Flag: is this a national series */
  boolean national;
  /** Name of the tournament */
  String tournament;
  /** Name of the series */
  String name;
  /** Character set, e.g., "utf-8" */
  String charset = "utf-8";
  /** Teams in the series */
  Vector<Team> teams = new Vector<Team> ();
  /** Competitors in the series */
  Vector<Competitor> competitors = new Vector<Competitor> ();
  /** Judges in the series */
  Vector<Judge> judges = new Vector<Judge> ();
  /** Flag: has the series been modified */
  private transient boolean modified = false;
  /** Contests in the series */
  private Vector<Contest> contests = new Vector<Contest> ();
  /** The countries */
  public Countries countries;
  /** The locale for this series */
  Locale locale;

  /** Constructor
   * @param locale	Locale of the series
   */
  public Series (Locale locale)
  {
    countries = new Countries (this.locale = locale);
    tournament = Constraints.CONTEST_TITLE;
  }

  /** Convert the series to a printable string */
  public String toString () { return tournament + " - " + name; }

  /** Flag the series modified */
  public void flagModified () { modified = true; }
  /** Determine whether the series data has been modified */
  public boolean isModified () { return modified; }

  /** Flag all contests in the series modified */
  public void flagContestsModified () {
    for (Contest contest : contests)
      contest.flagModified ();
  }

  /** Change the locale for this series
   * @param locale	the desired locale
   */
  public void setLocale (Locale locale)
  {
    if (this.locale != locale) {
      modified = true;
      countries.setLocale (this.locale = locale);
    }
  }

  /** Get the name of the series */
  public String getName () { return name; }
  /** Set the name of the series */
  public void setName (String name) {
    this.name = name;
    modified = true;
  }
  /** Get the name of the tournament */
  public String getTournament () { return tournament; }
  /** Get the number of judges */
  public int getNumJudges () { return judges.size (); }
  /** Get the nth judge */
  public Judge getJudge (int i) { return judges.elementAt (i); }
  /** Get a collection of the judges */
  public Collection<Judge> getJudges () { return judges; }
  /** Get the number of competitors */
  public int getNumCompetitors () {
    return competitors.size ();
  }
  /** Get the nth competitor */
  public Competitor getCompetitor (int i) {
    return competitors.elementAt (i);
  }
  /** Get a collection of the competitors */
  public Collection<Competitor> getCompetitors () { return competitors; }
  /** Get the number of teams */
  public int getNumTeams () {
    return teams.size ();
  }
  /** Get the nth team */
  public Team getTeam (int i) {
    return teams.elementAt (i);
  }
  /** Get a collection of all teams */
  public Collection<Team> getTeams () { return teams; }
  /** Get the number of contests */
  public int getNumContests () {
    return contests.size ();
  }
  /** Get the nth contest */
  public Contest getContest (int i) {
    return contests.elementAt (i);
  }
  /** Get a collection of all contests */
  public Collection<Contest> getContests () { return contests; }

  /** See if this is a national series */
  public boolean isNational () { return national; }
  /** Toggle the "national" status of the series
   * @return	the new status
   */
  public boolean toggleNational () {
    flagContestsModified ();
    return national = !national;
  }

  /** Get the locale for this series
   * @return	the locale
   */
  public Locale getLocale () { return locale; }

  /** Update the name of the series according to a file
   * @param file	the file
   */
  private void setName (File file)
  {
    this.file = file;
    name = file.getName ();
    int i = name.lastIndexOf (FILE_EXTENSION);
    if (i >= 0)
      name = name.substring (0, i);
  }

  /** Read the series from permanent storage
   * @param file	the input file
   * @exception		FileNotFoundException if the file cannot be found
   * @exception		IOException if the file cannot be read
   */
  public void read (File file)
    throws FileNotFoundException, IOException
  {
    setName (file);
    read (new FileReader (file));
  }

  /** Write the series to permanent storage
   * @param file	the output file
   * @exception		IOException if the file cannot be read
   */
  public void write (File file)
    throws IOException
  {
    write (new FileWriter (file));
    setName (file);
  }

  /** Read the series from permanent storage
   * @param reader	the input stream
   * @exception		IOException if the file cannot be read
   */
  private void read (Reader reader)
    throws IOException
  {
    LineNumberReader in = new LineNumberReader (reader);
    String line = in.readLine ();
    try {
      int i = line.indexOf (' ');
      if (i >= 0 && line.substring (0, i).equals ("national"))
	national = true;
      else if (i >= 0 && line.substring (0, i).equals ("international"))
	national = false;
      else
	throw new Exception ();
      setLocale (new Locale (line.substring (i + 1), ""));
      line = in.readLine ();
      if (line == null) throw new Exception ();
      tournament = line;
      line = in.readLine ();
      if (line.startsWith ("min ")) {
	/* ignore the line for backward compatibility */
	line = in.readLine ();
      }
      if (line.startsWith ("max ")) {
	/* ignore the line for backward compatibility */
	line = in.readLine ();
      }
      if (line.startsWith ("rect ")) {
	/* ignore the line for backward compatibility */
	line = in.readLine ();
      }
      if (line.startsWith ("charset ")) {
	charset = line.substring (8);
	line = in.readLine ();
      }
      if (line.equals ("judges")) {
	judges = new Vector<Judge> ();
	for (;;) {
	  line = in.readLine ();
	  i = line.indexOf (' ');
	  if (i < 0) break;
	  Country c = countries.get (line.substring (0, 2));
	  if (c == null) throw new Exception ("no such country");
	  add (new Judge (line.substring (i + 1), c));
	}
      }
      else
	throw new Exception ();
      if (line.equals ("competitors")) {
	competitors = new Vector<Competitor> ();
	for (;;) {
	  line = in.readLine ();
	  i = line.indexOf (' ');
	  if (i < 0) break;
	  Country c = countries.get (line.substring (0, 2));
	  if (c == null) throw new Exception ("no such country");
	  add (new Competitor (line.substring (i + 1), c));
	}
      }
      else
	throw new Exception ();
      if (line.equals ("teams")) {
	teams = new Vector<Team> ();
	for (;;) {
	  line = in.readLine ();
	  i = line.indexOf (' ');
	  if (i < 0) break;
	  int size = Integer.parseInt (line.substring (0, i));
	  Team team = new Team (this, null);
	  while (size-- > 0) {
	    int j = line.indexOf (' ', i + 1);
	    int num = Integer.parseInt (line.substring (i + 1, j));
	    i = j;
	    if (num < 0 || num >= competitors.size ())
	      throw new Exception ("no competitor " + num);
	    Competitor comp = competitors.elementAt (num);
	    if (team.competitors.contains (comp))
	      throw new Exception ("duplicate competitor " + num);
	    if (comp.team != null)
	      throw new Exception ("duplicate team for competitor " + num);
	    comp.team = team;
	    team.competitors.addElement (comp);
	  }
	  team.name = line.substring (i + 1);
	  add (team);
	}
      }
      else
	throw new Exception ();
      if (line.equals ("contests")) {
	contests = new Vector<Contest> ();
	for (;;) {
	  line = in.readLine ();
	  if (line == null) break;
	  i = line.indexOf (' ');
	  if (i < 0) break;
	  int numComp = Integer.parseInt (line.substring (0, i));
	  int j = line.indexOf (' ', i + 1);
	  int numJudges = Integer.parseInt (line.substring (i + 1, j));
	  Vector<Judge> jury = new Vector<Judge> (numJudges);
	  i = j;
	  for (; numJudges-- > 0; i = j) {
	    j = line.indexOf (' ', i + 1);
	    int num = Integer.parseInt (line.substring (i + 1, j));
	    if (num < 0 || num > judges.size ())
	      throw new Exception ("no judge " + num);
	    jury.addElement (judges.elementAt (num));
	  }
	  final String format = line.substring (i + 1);
	  Contest contest =
	    new Contest (line = in.readLine (), jury);
	  contest.setFormat (format);
	  add (contest);
	  while (numComp-- > 0) {
	    line = in.readLine ();
	    i = line.indexOf (' ');
	    int num = Integer.parseInt (line.substring (0, i));
	    if (num < 0 || num > competitors.size ())
	      throw new Exception ("no competitor " + num);
	    Competitor comp = competitors.elementAt (num);
	    j = line.indexOf (' ', i + 1);
	    num = Integer.parseInt (line.substring (i + 1, j));
	    Mirror mirror = new Mirror (comp, contest, num);
	    i = j;
	    j = line.indexOf (' ', i + 1);
	    mirror.penalty = Integer.parseInt (line.substring (i + 1, j));
	    if (mirror.penalty != 0) {
	      i = j;
	      j = line.indexOf (' ', i + 1);
	      mirror.penaltyReason = line.charAt (j - 1);
	    }
	    Integer points[] = new Integer[jury.size ()];
	    for (numJudges = 0; numJudges < jury.size (); numJudges++) {
	      i = j;
	      j = line.indexOf (' ', i + 1);
	      if (line.charAt (j - 1) != '*') {
		int pt = Integer.parseInt (line.substring (i + 1, j));
		if (pt < Constraints.MIN_POINTS || pt > Constraints.MAX_POINTS)
		  System.err.println (name + ":" + in.getLineNumber () +
			     ": score " + pt + " out of range, ignoring");
		else
		  points[numJudges] = new Integer (pt);
	      }
	    }
	    mirror.setPoints (points);
	    contest.add (mirror);
	  }
	}
	if ((line = in.readLine ()) != null)
	  throw new Exception ();
      }
      else
	throw new Exception ();
    }
    catch (IOException e) {
      throw e;
    }
    catch (Exception e) {
      String msg = e.toString ();
      if (msg == null)
	msg = "unrecognized input";
      if (line == null)
	line = "(end of file)";
      else
	line = "'" + line + "'";
      throw new IOException (name + ":" + in.getLineNumber () +
			     ": " + msg + ": " + line);
    }
    modified = false;
  }

  /** Write the series to permanent storage
   * @param writer	the output stream
   */
  private void write (Writer writer)
  {
    PrintWriter out = new PrintWriter (writer);
    int i;
    out.print (national ? "national " : "international ");
    out.println (locale.getLanguage ());
    out.println (tournament);
    out.println ("charset " + charset);
    out.println ("judges");
    for (Judge judge : judges)
      out.println (judge.country.code + " " + judge.name);
    out.println ("competitors");
    for (Competitor competitor : competitors)
      out.println (competitor.country.code + " " + competitor.name);
    out.println ("teams");
    for (Team team : teams) {
      out.print (team.size ());
      for (Competitor competitor : team.competitors)
	out.print (" " + competitors.indexOf (competitor));
      out.print (" ");
      out.println (team.name);
    }
    out.println ("contests");
    for (Contest contest : contests) {
      out.print (contest.mirrors.size ());
      out.print (" ");
      out.print (contest.judges.size ());
      for (Judge judge : contest.judges)
	out.print (" " + judges.indexOf (judge));
      out.print (" ");
      out.println (contest.getFormat () == null ? "" : contest.getFormat ());
      out.println (contest.name);
      for (Mirror mirror : contest.mirrors) {
	out.print (competitors.indexOf (mirror.competitor) + " " +
		   mirror.number + " " + mirror.penalty);
	if (mirror.penalty != 0)
	  out.print (" " + mirror.penaltyReason);
	Integer points[] = mirror.getPoints ();
	for (i = 0; i < points.length; i++)
	  out.print (points[i] == null ? " *" : " " + points[i]);
	out.println (" ");
      }
    }
    out.close ();
    modified = false;
  }

  /** Add a judge to the series
   * @param judge	the judge to be added
   */
  public void add (Judge judge) {
    judge.series = this;
    judges.add (judge);
    modified = true;
  }

  /** Add a competitor to the series
   * @param competitor	the competitor to be added
   */
  public void add (Competitor competitor) {
    competitor.series = this;
    competitors.add (competitor);
    modified = true;
  }

  /** Add a team to the series
   * @param team	the team to be added
   */
  public void add (Team team) {
    team.series = this;
    teams.add (team);
    modified = true;
  }

  /** Add a contest to the series
   * @param contest	the contest to be added
   */
  public void add (Contest contest) {
    contest.series = this;
    contests.add (contest);
    modified = true;
  }

  /** Remove a judge from the series
   * @param judge	the judge to be removed
   * @return		true if the judge belonged to the series
   */
  public boolean remove (Judge judge)
  {
    if (!judges.remove (judge))
      return false;
    for (Contest contest : contests)
      contest.remove (judge);
    return modified = true;
  }

  /** Remove a competitor from the series
   * @param competitor	the competitor to be removed
   * @return		true if the competitor belonged to the series
   */
  public boolean remove (Competitor competitor)
  {
    if (!competitors.remove (competitor))
      return false;
    if (competitor.team != null)
      competitor.team.remove (competitor);
    for (Contest contest : contests)
      contest.remove (competitor);
    return modified = true;
  }

  /** Remove a team from the series
   * @param team	the team to be removed
   * @return		true if the team belonged to the series
   */
  public boolean remove (Team team)
  {
    if (!teams.remove (team))
      return false;
    team.clear ();
    return modified = true;
  }

  /** Remove a contest from the series
   * @param contest	the contest to be removed
   * @return		true if the contest belonged to the series
   */
  public boolean remove (Contest contest)
  {
    if (!contests.remove (contest))
      return false;
    return modified = true;
  }

  /** Generate the HTML output
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generate (ErrorReporter reporter)
    throws IOException
  {
    File dir = mkdir (file.getParentFile (), name);
    HTMLWriter html = new HTMLWriter (dir, "index.html", "..", locale);
    html.header (tournament, name, "series.index", charset);

    html.out.print ("<h2>");
    html.link ("mirrors.index", "mirrors.html", null, null);
    html.out.println ("</h2>");

    html.out.print ("<h2>");
    html.link ("results.index", "results.html", null, null);
    html.out.println ("</h2>");

    html.out.print ("<h2>");
    html.link ("teams.index", "teams.html", null, null);
    html.out.println ("</h2>");

    html.out.print ("<h2>");
    html.link ("awards.index", "awards.html", null, null);
    html.out.println ("</h2>");

    html.footer ();

    generateMirrors (dir, reporter);
    generateResults (dir, reporter);
    generateAwards (dir, 5);
    generateTeams (dir);
  }

  /** Generate an index of mirror numbers in the series
   * @param dir		the directory where the index is to be created
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateMirrors (File dir, ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "mirrors.html", "..", locale);
    html.header (tournament, name, "mirrors.index", charset);
    // flag: has anything been printed?
    boolean output = false;

    for (Contest contest : contests) {
      if (!output) {
	output = true;
	html.out.println ("<ol>");
      }
      html.out.print ("<li>");
      html.link (null, "mirrors.html", contest.name, null);
      html.out.println ("</li>");
      contest.generateMirrors (mkdir (dir, contest.name), charset, reporter);
    }

    if (output)
      html.out.println ("</ol>");
    html.footer ();
  }

  /** Generate individual results for the series
   * @param dir		the directory where the file is to be created
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateResults (File dir, ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "results.html", "..", locale);
    html.header (tournament, name, "results.index", charset);

    html.out.println ("<table class=\"results\" frame=\"hsides\" rules=\"groups\">");
    html.out.println ("<colgroup align=\"right\" span=\"1\">");
    html.out.println ("<colgroup align=\"left\" span=\"2\">");
    html.out.println ("<colgroup align=\"right\" span=\"1\">");
    html.out.println ("<colgroup align=\"right\" span=\"" +
		      2 * contests.size () + "\">");
    html.out.println ("<thead><tr>");
    html.tag ("th", "results.rank", "scope=\"col\"");
    html.tag ("th", "results.name", "scope=\"col\"");
    html.tag ("th", "results.team", "scope=\"col\"");
    html.tag ("th", "results.total", "scope=\"col\"");

    for (Contest contest : contests) {
      html.out.print ("<th scope=\"col\" colspan=\"2\">");
      html.link (null, "results.html", contest.name, null);
      html.out.println ("</th>");
      contest.generateResults (mkdir (dir, contest.name), charset, reporter);
    }
    html.out.println ("</tr></thead>\n<tbody>");

    // Display the rankings
    int place = 0, rank = 0, lasttotal = -1;
    for (Competitor competitor : getRanking ()) {
      int total = getPoints (competitor);
      if (total != lasttotal) {
	rank = place + 1;
	lasttotal = total;
      }

      html.out.println ("<tr>\n <td scope=\"row\" class=\"rank\">" +
			rank + "</td>\n <td>" +
			competitor.name + "</td>\n <td>" +
			(competitor.team == null
			 ? competitor.country.name
			 : competitor.team.name + ", " +
			 competitor.country.name) +
			"</td>\n <td class=\"num\">" + total + "</td>");

      for (Contest contest : contests) {
	Mirror mirror = contest.getMirror (competitor);
	if (mirror != null) {
	  html.out.print ("<td class=\"mirrornumber\">" + mirror.rank + ": ");
	  html.link (mirror.getTotalRectified (), "results.html",
		     contest.name, "m" + mirror.number);
	  html.out.println ("</td><td class=\"mirror\">");
	  if (contest.getFormat () != null) {
	    html.out.print (' ');
	    html.img (contest.name, new Integer (mirror.number).toString (),
		      contest.getFormat ());
	  }
	  html.out.println ("</td>");
	}
	else
	  html.out.println ("<td class=\"mirrornumber\" colspan=\"2\"></td>");
      }

      html.out.println ("</tr>");
      place++;
    }

    if (place == 0)
      html.out.println ("<tr><td></td></tr>"); // thead, tbody mandate tr
    html.out.println ("</tbody>\n</table>");
    html.footer ();
  }

  /** Generate award list for the series
   * @param dir		the directory where the file is to be created
   * @param top		number of top ranks to list
   * @exception		IOException if the output cannot be created
   */
  public void generateAwards (File dir, int top)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "awards.html", "..", locale);
    html.header (tournament, name, "awards.index", charset);

    html.out.println ("<h2>" + name + "</h2>\n");

    for (Contest contest : contests) {
      html.out.print ("<h3>" + name + "/" + contest.name + "</h3>\n");
      contest.generateAwards (html, top);
    }

    html.tag ("h3", "awards.total", null);

    // Display the rankings
    int place = 0, rank = 0, lasttotal = -1;
    for (Competitor competitor : getRanking ()) {
      place++;

      int total = getPoints (competitor);
      if (total != lasttotal) {
	rank = place;
	lasttotal = total;
      }

      if (rank > top)
	break;

      html.out.print ("<p>" + rank + "." +
		      competitor.name + "; " +
		      (competitor.team == null
		       ? competitor.country.name
		       : competitor.team.name + ", " +
		       competitor.country.name) + "(");

      for (Enumeration<Contest> c = contests.elements ();
	   c.hasMoreElements (); ) {
	Contest contest = c.nextElement ();
	Mirror mirror = contest.getMirror (competitor);
	if (mirror != null)
	  html.out.print (mirror.getTotalRectified ());
	else
	  html.out.print ("?");
	if (c.hasMoreElements ())
	  html.out.print ("+");
      }

      html.out.println ("=" + total + ")</p>");
    }

    html.footer ();
  }

  /** Generate team results for the series
   * @param dir		the directory where the file is to be created
   * @exception		IOException if the output cannot be created
   */
  public void generateTeams (File dir)
    throws IOException
  {
    /** Minimum number of participants in a team */
    int numCompetitors = Integer.MAX_VALUE;
    for (int i = teams.size (); i-- > 0; ) {
      int size = teams.elementAt (i).size ();
      if (size < numCompetitors)
	numCompetitors = size;
    }

    HTMLWriter html = new HTMLWriter (dir, "teams.html", "..", locale);
    html.header (tournament, name, "teams.index", charset);

    html.out.println ("<table class=\"results\">");
    html.out.println ("<thead><tr>");
    html.tag ("th", "results.rank", "scope=\"col\"");
    html.tag ("th", "results.team", "scope=\"col\"");
    html.tag ("th", "results.total", "scope=\"col\"");
    for (Enumeration<Contest> c = contests.elements ();
	 c.hasMoreElements (); ) {
      Contest contest = c.nextElement ();
      html.out.print ("<th scope=\"col\">");
      html.link (null, "results.html", contest.name, null);
      html.out.println ("</th>");
    }
    html.out.println ("</tr></thead>\n<tbody>");

    // Display the rankings
    int place = 0, rank = 0, lasttotal = -1;
    for (Object[] data : getTeamRanking (numCompetitors)) {
      Team team = (Team) data[0];
      int points[] = (int[]) data[1];

      if (points[0] != lasttotal) {
	rank = place + 1;
	lasttotal = points[0];
      }

      html.out.print ("<tr>\n<td class=\"rank\">" + rank + "</td>\n<td>");
      html.html (team.name);
      html.out.println ("</td>\n<td class=\"score\">" + points[0] + "</td>");

      {
	int i = 0;
	for (Contest contest : contests)
	  html.out.println ("<td class=\"score\">" + points[++i] + "</td>");
      }

      html.out.println ("</tr>");
      for (Competitor competitor : team.competitors) {
	html.out.print ("<tr>\n<td></td>\n<td>");
	html.html ("&nbsp;&nbsp;" + competitor.name);
	html.out.println ("<td class=\"score\">" + getPoints (competitor) +
			  "</td>");
	for (Contest contest : contests) {
	  Mirror mirror = contest.getMirror (competitor);
	  if (mirror != null) {
	    html.out.print ("<td>" + mirror.rank + ": ");
	    html.link (mirror.getTotalRectified (), "results.html",
		       contest.name, Integer.toString (mirror.number));
	    if (contest.getFormat () != null) {
	      html.out.print (' ');
	      html.img (contest.name, new Integer (mirror.number).toString (),
			contest.getFormat ());
	    }
	    html.out.println ("</td>");
	  }
	  else
	    html.out.println ("<td></td>");
	}
      }
      html.out.println ("</tr>");
      place++;
    }

    if (place == 0)
      html.out.println ("<tr><td></td></tr>"); // thead, tbody mandate tr
    html.out.println ("</tbody>\n</table>");
    html.footer ();
  }

  /** Determine the rankings for the series
   * @return	Competitors in the ranking order
   */
  public List<Competitor> getRanking ()
  {
    List<Competitor> ranking = new ArrayList<Competitor> (competitors);
    final Collator collator = Collator.getInstance (locale);
    Collections.sort (ranking, new Comparator<Competitor> () {
	public int compare (Competitor a, Competitor b) {
	  int aTotal = getPoints (a), bTotal = getPoints (b);
	  if (aTotal != bTotal)
	    return bTotal - aTotal;
	  /* If the total points are even, the last contest where at
	  least one is participating will determine the ranking. */
	  for (int i = contests.size (); i-- > 0; ) {
	    Contest last = contests.elementAt (i);
	    Mirror aMirror = last.getMirror (a), bMirror = last.getMirror (b);
	    if (aMirror == null && bMirror == null)
	      continue;
	    if (aMirror == null)
	      return -1;
	    if (bMirror == null)
	      return 1;
	    if (aMirror.rank != bMirror.rank)
	      return aMirror.rank - bMirror.rank;
	    aTotal = aMirror.getTotalRectified ();
	    bTotal = bMirror.getTotalRectified ();
	    if (aTotal != bTotal)
	      return bTotal - aTotal;
	  }

	  // It is an even game: collate the competitors in alphabetical order
	  return collator.compare (a.name, b.name);
	}
      });
    return ranking;
  }

  /** Determine the rankings for the teams
   * @param numCompetitors	number of team members to consider
   * @return	Vector of Object[] (Team,int[])
   *		where the first int are the total points
   *		and subsequent ints are the points for each contest
   */
  public Vector<Object[]> getTeamRanking (int numCompetitors)
  {
    Vector<Object[]> tr = new Vector<Object[]> ();
    for (Team team : teams) {
      int points[] = getPoints (team, numCompetitors);
      int i;
      for (i = 0; i < tr.size (); i++)
	if (points[0] > ((int[]) (tr.elementAt (i)[1]))[0])
	  break;
      Object obj[] = new Object[] {team, points};
      if (i == tr.size ())
	tr.addElement (obj);
      else
	tr.insertElementAt (obj, i);
    }
    return tr;
  }

  /** Determine the total points of a contestant in this series
   * @param competitor		the contestant
   * @return			total number of rectified points minus penalty
   */
  public int getPoints (Competitor competitor)
  {
    int total = 0;
    for (Contest contest : contests) {
      Mirror mirror = contest.getMirror (competitor);
      if (mirror != null)
	total += mirror.getTotalRectified ();
    }
    return total;
  }

  /** Determine the total points of a team in this series
   * @param team		the team
   * @param numCompetitors	number of team members to consider
   * @return			total number of rectified points minus penalty,
   *				plus total points in each contest
   */
  public int[] getPoints (Team team, int numCompetitors)
  {
    int total[] = new int[1 + contests.size ()];
    for (int c = contests.size (); c-- > 0; ) {
      Contest contest = contests.elementAt (c);
      int points[] = new int[numCompetitors];
      for (Competitor competitor : team.competitors) {
	final Mirror mirror = contest.getMirror (competitor);
	final int p = mirror == null ? 0 : mirror.getTotalRectified ();
	for (int i = 0; i < numCompetitors; i++) {
	  if (points[i] < p) {
	    for (int j = numCompetitors; --j > i; )
	      points[j] = points[j - 1];
	    points[i] = p;
	    break;
	  }
	}
      }
      for (int i = numCompetitors; i-- > 0; )
	total[1 + c] += points[i];
      total[0] += total[1 + c];
    }
    return total;
  }

  /** Create a directory
   * @param parent	the parent directory
   * @param directory	the directory name
   * @exception		IOException if the directory cannot be created
   * @return		a file object to the directory
   */
  public static File mkdir (File parent, String directory)
    throws IOException
  {
    File file = new File (parent, directory);
    if (!file.isDirectory () && !file.mkdir ())
      throw new IOException ("mkdir " + file);
    return file;
  }
}
