package results.ui;

import results.Team;
import results.Competitor;

import java.util.Locale;
import javax.swing.DefaultCellEditor;
import javax.swing.JTable;
import javax.swing.event.TableModelEvent;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

/** Table for viewing and editing team data */
final class TeamTable
  extends JTable
{
  /** The serialization version */
  private static final long serialVersionUID = 1L;
  /** The series dialog containing this table */
  private SeriesDialog parent;
  /** The team name cell renderer */
  private DefaultTableCellRenderer rteam;
  /** The competitor cell renderer */
  private DefaultTableCellRenderer rcompetitor;

  /** The competitor box */
  private CompetitorBox competitorbox;

  /** The team name column number */
  private static final int COLUMN_TEAM = 0;
  /** The first column for competitors */
  private static final int COLUMN_COMPETITORS = 1;
  /** Maximum number of members in a team */
  private int maxMembers;

  /** Constructor
   * @param dialog	the containing series dialog
   */
  TeamTable (SeriesDialog dialog)
  {
    parent = dialog;
    setModel (new AbstractTableModel ()
      {
	/** The serialization version */
	private static final long serialVersionUID = 1L;
	public int getColumnCount () { return 2 + TeamTable.this.maxMembers; }
	public int getRowCount () {
	  return parent == null ? 1 : parent.series.getNumTeams () + 1;
	}
	public boolean isCellEditable (int row, int col) {
	  return true;
	}
	public Class<?> getColumnClass (int col) {
	  switch (col) {
	  case COLUMN_TEAM:
	    return String.class;
	  default:
	    return Competitor.class;
	  }
	}
	public Object getValueAt (int row, int col) {
	  final int rows = getRowCount ();
	  if (row < rows - 1) {
	    Team team = parent.series.getTeam (row);
	    switch (col) {
	    case COLUMN_TEAM:
	      return team.getName ();
	    default:
	      col -= COLUMN_COMPETITORS;
	      if (col < team.size ())
		return team.elementAt (col);
	    }
	  }
	  return null;
	}
	public void setValueAt (Object value, int row, int col) {
	  final int rows = getRowCount ();
	  Team team;
	  if (row < rows - 1)
	    team = parent.series.getTeam (row);
	  else {
	    team = new Team (null, "");
	    parent.series.add (team);
	  }
	  switch (col) {
	  case COLUMN_TEAM:
	    if (value == null || value.equals ("")) {
	      if (row >= rows - 1 ||
		  parent.confirm ("team.remove", team.getName ()))
		parent.series.remove (team);
	    }
	    else
	      team.setName ((String) value);
	    break;
	  default:
	    col -= COLUMN_COMPETITORS;
	    if (col < team.size ()) {
	      if (value == null) {
		team.removeElementAt (col);
		TeamTable.this.computeMaxMembers ();
	      }
	      else
		team.setElementAt ((Competitor) value, col);
	    }
	    else if (value != null) {
	      team.add ((Competitor) value);
	      TeamTable.this.computeMaxMembers ();
	    }
	  }
	}
	public String getColumnName (int col) {
	  switch (col) {
	  case COLUMN_TEAM:
	    return parent.translator.lookup ("team.nTeam");
	  default:
	    return parent.translator.lookup ("team.nCompetitor");
	  }
	}
      });
    setDefaultEditor (Competitor.class, new DefaultCellEditor
      (competitorbox = new CompetitorBox (parent.series)));
    setDefaultRenderer (String.class,
			rteam = new DefaultTableCellRenderer ());
    setDefaultRenderer (Competitor.class,
			rcompetitor = new DefaultTableCellRenderer ());
  }

  /** Set the locale
   * @param locale	the new locale
   */
  public void setLocale (Locale locale)
  {
    super.setLocale (locale);
    if (parent == null)
      return;
    if (rteam != null)
      rteam.setToolTipText (parent.translator.lookup ("team.tTeam"));
    if (rcompetitor != null)
      rcompetitor.setToolTipText (parent.translator.lookup
				  ("team.tCompetitor"));
    tableChanged (new TableModelEvent
		  (getModel (), TableModelEvent.HEADER_ROW));
  }

  /** Determine the maximum number of members in any team */
  void computeMaxMembers ()
  {
    maxMembers = 0;
    for (Team team : parent.series.getTeams ()) {
      int members = team.size ();
      if (members > maxMembers)
	maxMembers = members;
    }
    tableChanged (new TableModelEvent
      (getModel (), TableModelEvent.HEADER_ROW));
  }

  /** Update the data on available competitors */
  public void updateCompetitors () {
    if (competitorbox != null)
      competitorbox.updateCompetitors (parent.series);
    computeMaxMembers ();
  }
}
